"""Arkprts errors."""
import json
import typing


class BaseArkprtsError(Exception):
    """Base class for all Arkprts errors."""

    message: str = "Arkprts error."

    def __init__(self, message: typing.Optional[str] = None) -> None:
        super().__init__(message or self.message)


class NotLoggedInError(BaseArkprtsError):
    """Raised when a user is not logged in."""

    message = "Not logged in."


class ArkPrtsError(BaseArkprtsError):
    """Raised when result code is not 0."""

    data: typing.Mapping[str, typing.Any]

    def __init__(self, data: typing.Mapping[str, typing.Any]) -> None:
        self.data = data
        super().__init__(f"[{data['result']}] {self.message} {data}")


class GameServerError(BaseArkprtsError):
    """Game server error."""

    status_code: int
    error: str
    code: int
    msg: str
    info: typing.Mapping[str, typing.Any]

    def __init__(self, data: typing.Mapping[str, typing.Any]) -> None:
        self.status_code = data.get("statusCode", 400)
        self.error = data["error"]
        self.code = data.get("code", 0)
        self.msg = data.get("msg", "")
        self.info = json.loads(data.get("info", "{}"))

        super().__init__(str(data))


class GeetestError(ArkPrtsError):
    """Raised when login is flagged by geetest."""

    message: str = "Geetest verification is required."
    challenge: str
    gt: str

    def __init__(self, data: typing.Mapping[str, typing.Any]) -> None:
        self.challenge = data["captcha"]["challenge"]
        self.gt = data["captcha"]["gt"]
        super().__init__(data)


class InvalidStatusError(BaseArkprtsError):
    """Raised when a response has an invalid status code."""

    status: int
    data: typing.Mapping[str, typing.Any]

    def __init__(self, status: int, data: typing.Mapping[str, typing.Any]) -> None:
        self.status = status
        self.data = data

        super().__init__(f"[{status}] {data}")


class InvalidContentTypeError(BaseArkprtsError):
    """Raised when a response has an invalid content type."""

    message = "Invalid content type."
    content: str

    def __init__(self, content: str) -> None:
        self.content = content
        super().__init__()



"""Arknights API social models.

Any field description prefixed with IDK means it's just a guess.
"""
import typing

import pydantic

from . import base


class Avatar(base.BaseModel):
    """User display avatar."""

    type: typing.Literal["ASSISTANT", "ICON", "DEFAULT"] = "DEFAULT"
    """Avatar type."""
    id: typing.Optional[str] = None
    """Avatar ID. For example a skin ID."""


class Skill(base.BaseModel):
    """Skill of a character."""

    skill_id: str = pydantic.Field(alias="skillId")
    """Skill ID."""
    unlock: bool
    """Whether the skill is unlocked."""
    state: bool = pydantic.Field(repr=False)
    """IDK. Always false."""
    specialize_level: int = pydantic.Field(alias="specializeLevel")
    """Skill mastery level."""
    complete_upgrade_time: typing.Optional[base.ArknightsTimestamp] = pydantic.Field(alias="completeUpgradeTime")
    """IDK. Time left until skill upgrade is complete. Is raw -1 if not upgrading."""

    @property
    def static(self) -> base.DDict:
        """Static data for this skill."""
        return self.client.gamedata.skill_table[self.skill_id]


class UniEquip(base.BaseModel):
    """Equipped modules."""

    hide: bool
    """IDK. Whether the module is publicly hidden."""
    locked: bool
    """Whether module access is locked."""
    level: int
    """Module level."""


class AssistChar(base.BaseModel):
    """Publicly visible operator info."""

    char_id: str = pydantic.Field(alias="charId")
    """Character ID."""
    skin_id: str = pydantic.Field(alias="skinId")
    """Equipped skin ID."""
    skills: typing.Sequence[Skill]
    """Operator skills."""
    main_skill_lvl: int = pydantic.Field(alias="mainSkillLvl")
    """Level of the equipped skill."""
    skill_index: int = pydantic.Field(alias="skillIndex")
    """Index of the equipped skill."""
    evolve_phase: int = pydantic.Field(alias="evolvePhase")
    """Elite phase."""
    favor_point: int = pydantic.Field(alias="favorPoint")
    """Operator trust points."""
    potential_rank: int = pydantic.Field(alias="potentialRank")
    """Operator potential. Starts at 0."""
    level: int
    """Operator level."""
    crisis_record: typing.Mapping[str, int] = pydantic.Field(alias="crisisRecord")
    """Maximum risk this operator was used as a support in. -1 if never used."""
    current_equip: typing.Optional[str] = pydantic.Field(alias="currentEquip")
    """ID of the currently equipped module."""
    equip: typing.Mapping[str, UniEquip]
    """Equipped modules. Module ID to module info."""
    tmpl: typing.Mapping[str, base.DDict] = pydantic.Field(default_factory=base.DDict, repr=False)
    """Alternative operator class data. Only for Amiya."""

    @property
    def static(self) -> base.DDict:
        """Static data for this operator."""
        return self.client.gamedata.character_table[self.char_id]

    @property
    def trust(self) -> int:
        """Trust calculated from favor_point."""
        return self.client.gamedata.calculate_trust_level(self.favor_point)


class PlacedMedal(base.BaseModel):
    """A single medal on a board."""

    id: str
    """Medal ID."""
    pos: typing.Tuple[int, int]
    """Medal position on the board."""

    @property
    def static(self) -> base.DDict:
        """Static data for this medal."""
        return self.client.gamedata.medal_table.medal_list[self.id]


class MedalBoardCustom(base.BaseModel):
    """Custom medal board layout."""

    layout: typing.Sequence[PlacedMedal]
    """Medals on the board."""


class MedalBoardTemplate(base.BaseModel):
    """Template medal board layout."""

    group_id: str = pydantic.Field(alias="groupId")
    """Medal board template ID."""
    medal_list: typing.Sequence[str] = pydantic.Field(alias="medalList")
    """Medal IDs on the board."""

    @property
    def static(self) -> base.DDict:
        """Static data for this medal board."""
        groups = self.client.gamedata.medal_table.medal_type_data.activity_medal.group_data
        return next(i for i in groups if i.group_id == self.group_id)


class MedalBoard(base.BaseModel):
    """Medal board info."""

    type: typing.Literal["CUSTOM", "TEMPLATE", "EMPTY"]
    """Medal board layout type."""
    custom: typing.Optional[MedalBoardCustom]
    """Custom medal board layout."""
    template: typing.Optional[MedalBoardTemplate]
    """Template medal board layout."""


class PartialPlayer(base.BaseModel):
    """Partial player info from search."""

    nickname: str = pydantic.Field(alias="nickName")
    """Player nickname."""
    nick_number: str = pydantic.Field(alias="nickNumber")
    """Player nickname number after #."""
    uid: str
    """Player UID."""
    friend_num_limit: int = pydantic.Field(alias="friendNumLimit")
    """How many friend slots are open."""
    server_name: str = pydantic.Field(alias="serverName")
    """Server name. Should always be Terra."""
    level: int
    """Player level."""
    avatar_id: str = pydantic.Field(alias="avatarId", repr=False)
    """IDK. Always 0."""
    avatar: typing.Optional[Avatar] = None
    """Selected avatar."""
    assist_char_list: typing.Sequence[typing.Optional[AssistChar]] = pydantic.Field(alias="assistCharList")
    """Assist operator list."""
    last_online_time: base.ArknightsTimestamp = pydantic.Field(alias="lastOnlineTime")
    """Last online time."""
    medal_board: MedalBoard = pydantic.Field(alias="medalBoard")
    """Medal board."""


class Player(PartialPlayer):
    """Player info."""

    register_ts: base.ArknightsTimestamp = pydantic.Field(alias="registerTs")
    """Account creation time."""
    main_stage_progress: typing.Optional[str] = pydantic.Field(alias="mainStageProgress")
    """Current main story stage ID. None if completed."""
    char_cnt: int = pydantic.Field(alias="charCnt")
    """Number of operators owned."""
    furn_cnt: int = pydantic.Field(alias="furnCnt")
    """Number of furniture owned."""
    secretary: str
    """ID of the secretary operator."""
    secretary_skin_id: str = pydantic.Field(alias="secretarySkinId")
    """ID of the secretary operator's skin."""
    resume: str
    """Player display bio."""
    team_v2: typing.Mapping[str, int] = pydantic.Field(alias="teamV2")
    """Amount of characters owned in each faction."""
    board: typing.Sequence[str]
    """Factions with full trust. Shows up blue in-game."""
    info_share: typing.Optional[base.ArknightsTimestamp] = pydantic.Field(alias="infoShare")
    """IDK."""
    recent_visited: bool = pydantic.Field(alias="recentVisited")
    """IDK."""
    info_share_visited: typing.Optional[int] = pydantic.Field(None, alias="infoShareVisited")
    """IDK."""



"""Arknights API data models.

Any field description prefixed with IDK means it's just a guess.
"""
import typing

import pydantic

from . import base


class Avatar(base.BaseModel):
    """User display avatar."""

    type: typing.Literal["ASSISTANT", "ICON", "DEFAULT"] = "DEFAULT"
    """Avatar type."""
    id: typing.Optional[str] = None
    """Avatar ID. For example a skin ID."""


class Status(base.BaseModel):
    """General user data."""

    nickname: str = pydantic.Field(alias="nickName")
    """Player nickname."""
    nick_number: str = pydantic.Field(alias="nickNumber")
    """Player nickname number after #."""
    level: int
    """Player level."""
    exp: int
    """Player experience."""
    social_point: int = pydantic.Field(alias="socialPoint")
    """Credit shop credits."""
    gacha_ticket: int = pydantic.Field(alias="gachaTicket")
    """Amount of single headhunting permits."""
    ten_gacha_ticket: int = pydantic.Field(alias="tenGachaTicket")
    """Amount of 10-headhunting permits."""
    instant_finish_ticket: int = pydantic.Field(alias="instantFinishTicket")
    """Amount of expedited completion permits."""
    hgg_shard: int = pydantic.Field(alias="hggShard")
    """Amount of yellow distinction certificates."""
    lgg_shard: int = pydantic.Field(alias="lggShard")
    """Amount of green commendation certificates.."""
    recruit_license: int = pydantic.Field(alias="recruitLicense")
    """Amount of recruitment permit."""
    progress: int = pydantic.Field(repr=False)
    """IDK."""
    buy_ap_remain_times: int = pydantic.Field(alias="buyApRemainTimes", repr=False)
    """IDK. Ap refers to sanity."""
    ap_limit_up_flag: bool = pydantic.Field(alias="apLimitUpFlag", repr=False)
    """IDK. Ap refers to sanity."""
    uid: str
    """User ID."""
    flags: typing.Mapping[str, bool] = pydantic.Field(repr=False)
    """Completed stories."""
    ap: int
    """Current sanity. Actually not a true value."""
    max_ap: int = pydantic.Field(alias="maxAp")
    """Max sanity."""
    pay_diamond: int = pydantic.Field(alias="payDiamond")
    """Bough originium prime."""
    free_diamond: int = pydantic.Field(alias="freeDiamond")
    """Earned originium prime."""
    diamond_shard: int = pydantic.Field(alias="diamondShard")
    """Amount of orundum."""
    gold: int
    """Amount of LMD."""
    practice_ticket: int = pydantic.Field(alias="practiceTicket")
    """Amount of training permits."""
    last_refresh_ts: base.ArknightsTimestamp = pydantic.Field(alias="lastRefreshTs")
    """IDK. When sanity was last incremented."""
    last_ap_add_time: base.ArknightsTimestamp = pydantic.Field(alias="lastApAddTime")
    """IDK."""
    last_online_ts: base.ArknightsTimestamp = pydantic.Field(alias="lastOnlineTs")
    """IDK. When the player was last online."""
    main_stage_progress: typing.Optional[str] = pydantic.Field(alias="mainStageProgress")
    """Current main story stage ID. None if completed."""
    register_ts: base.ArknightsTimestamp = pydantic.Field(alias="registerTs")
    """Account creation time."""
    server_name: str = pydantic.Field(alias="serverName")
    """Server name. Should always be Terra."""
    avatar_id: str = pydantic.Field(alias="avatarId", repr=False)
    """IDK. Always 0."""
    resume: str
    """Player display bio."""
    friend_num_limit: int = pydantic.Field(alias="friendNumLimit")
    """How many friend slots are open."""
    secretary: str
    """ID of the secretary operator."""
    secretary_skin_id: str = pydantic.Field(alias="secretarySkinId")
    """ID of the secretary operator's skin."""
    global_voice_lan: typing.Optional[str] = pydantic.Field(None, alias="globalVoiceLan")
    """Default voice-over language."""
    avatar: Avatar
    """Selected avatar."""

    # fmt: off
    # optional:
    monthly_subscription_start_time: typing.Optional[base.ArknightsTimestamp] = pydantic.Field(None, alias="monthlySubscriptionStartTime")
    """When the monthly subscription started."""
    monthly_subscription_end_time: typing.Optional[base.ArknightsTimestamp] = pydantic.Field(None, alias="monthlySubscriptionEndTime")
    """When the monthly subscription is ending."""
    tip_monthly_card_expire: typing.Optional[base.ArknightsTimestamp] = pydantic.Field(None, alias="tipMonthlyCardExpire")
    """IDK."""
    # fmt: on

    @property
    def basic_item_inventory(self) -> typing.Mapping[str, int]:
        """Basic item inventory. These are not shown in the inventory object."""
        return {
            "SOCIAL_PT": self.social_point,
            "AP_GAMEPLAY": self.ap,
            "4001": self.gold,
            "4002": self.pay_diamond + self.free_diamond,
            "4003": self.diamond_shard,
            "4004": self.hgg_shard,
            "4005": self.lgg_shard,
            "5001": self.exp,
            "6001": self.practice_ticket,
            "7001": self.recruit_license,
            "7002": self.instant_finish_ticket,
            "7003": self.gacha_ticket,
            "7004": self.ten_gacha_ticket,
        }


class SquadSlot(base.BaseModel):
    """Squad slot."""

    char_inst_id: int = pydantic.Field(alias="charInstId")
    """Operator index."""
    skill_index: int = pydantic.Field(alias="skillIndex")
    """Index of chosen skill."""
    current_equip: typing.Optional[str] = None
    """Currently equipped module ID."""
    tmpl: typing.Mapping[str, base.DDict] = pydantic.Field(default_factory=base.DDict, repr=False)
    """Alternative operator class data. Only for Amiya."""


class Squads(base.BaseModel):
    """Operator squad data."""

    squad_id: str = pydantic.Field(alias="squadId")
    """Squad ID."""
    name: str
    """Squad name."""
    slots: typing.Sequence[typing.Optional[SquadSlot]]
    """Equipped operators."""


class Skill(base.BaseModel):
    """Operator skill data."""

    skill_id: str = pydantic.Field(alias="skillId")
    """Skill ID."""
    unlock: bool
    """Whether the skill is unlocked."""
    state: bool = pydantic.Field(repr=False)
    """IDK. Always false."""
    specialize_level: int = pydantic.Field(alias="specializeLevel")
    """Skill mastery level."""
    complete_upgrade_time: typing.Optional[base.ArknightsTimestamp] = pydantic.Field(alias="completeUpgradeTime")
    """IDK. Time left until skill upgrade is complete. Is raw -1 if not upgrading."""

    @property
    def static(self) -> base.DDict:
        """Static data for this skill."""
        return self.client.gamedata.skill_table[self.skill_id]


class Equip(base.BaseModel):
    """Operator module data."""

    hide: bool
    """Whether the module is hidden."""
    locked: bool
    """Whether the module is locked."""
    level: int
    """Module level."""


class Character(base.BaseModel):
    """Operator data."""

    inst_id: int = pydantic.Field(alias="instId")
    """Index of the operator."""
    char_id: str = pydantic.Field(alias="charId")
    """Operator ID."""
    favor_point: int = pydantic.Field(alias="favorPoint")
    """Operator trust points."""
    potential_rank: int = pydantic.Field(alias="potentialRank")
    """Operator potential. Starts at 0."""
    main_skill_lvl: int = pydantic.Field(alias="mainSkillLvl")
    """Operator skill level."""
    skin: str
    """Operator skin ID."""
    level: int
    """Operator level."""
    exp: int
    """Operator experience."""
    evolve_phase: int = pydantic.Field(alias="evolvePhase")
    """Elite phase."""
    default_skill_index: int = pydantic.Field(alias="defaultSkillIndex")
    """Index of the default skill."""
    skills: typing.Sequence[Skill]
    """Operator skills."""
    voice_lan: str = pydantic.Field(alias="voiceLan")
    """Operator voice-over language."""
    current_equip: typing.Optional[str] = pydantic.Field(alias="currentEquip")
    """Currently equipped module."""
    equip: typing.Mapping[str, Equip]
    """Operator modules."""
    star_mark: bool = pydantic.Field(False, alias="starMark")
    """Whether the operator is marked as favorite."""
    tmpl: typing.Mapping[str, base.DDict] = pydantic.Field(default_factory=base.DDict, repr=False)
    """Alternative operator class data. Only for Amiya."""

    @property
    def static(self) -> base.DDict:
        """Static data for this operator."""
        return self.client.gamedata.character_table[self.char_id]

    @property
    def trust(self) -> int:
        """Trust calculated from favor_point."""
        return self.client.gamedata.calculate_trust_level(self.favor_point)


class CharGroup(base.BaseModel):
    """Additional operator data."""

    favor_point: int = pydantic.Field(alias="favorPoint")
    """Operator trust points."""

    @property
    def trust(self) -> int:
        """Trust calculated from favor_point."""
        return self.client.gamedata.calculate_trust_level(self.favor_point)


class Troops(base.BaseModel):
    """Operator data."""

    cur_char_inst_id: int = pydantic.Field(alias="curCharInstId")
    """Amount of owned operators."""
    cur_squad_count: int = pydantic.Field(alias="curSquadCount")
    """Amount of squads. Should be always 4."""
    squads: typing.Mapping[int, Squads]
    """Squad data."""
    chars: typing.Mapping[int, Character]
    """Operator data."""
    char_group: typing.Mapping[str, CharGroup] = pydantic.Field(alias="charGroup")
    """Additional operator data."""
    char_mission: typing.Mapping[str, typing.Mapping[str, int]] = pydantic.Field(alias="charMission", repr=False)
    """IDK. Special operation missions."""
    addon: base.DDict = pydantic.Field(default_factory=base.DDict, repr=False)
    """IDK."""


class Skins(base.BaseModel):
    """Operator skin data."""

    character_skins: typing.Mapping[str, bool] = pydantic.Field(alias="characterSkins")
    """Owned skins."""
    skin_ts: typing.Mapping[str, base.ArknightsTimestamp] = pydantic.Field(alias="skinTs")
    """When the skins were obtained."""


class AssistChar(base.BaseModel):
    """Assist operator data."""

    char_inst_id: int = pydantic.Field(alias="charInstId")
    """Index of the operator."""
    skill_index: int = pydantic.Field(alias="skillIndex")
    """Index of the selected skill."""
    current_equip: typing.Optional[str] = pydantic.Field(alias="currentEquip")
    """Currently equipped module."""
    tmpl: typing.Mapping[str, base.DDict] = pydantic.Field(default_factory=base.DDict, repr=False)
    """Alternative operator class data. Only for Amiya."""


class Social(base.BaseModel):
    """Social data."""

    assist_char_list: typing.Sequence[typing.Optional[AssistChar]] = pydantic.Field(alias="assistCharList")
    """Support operators."""
    yesterday_reward: base.DDict = pydantic.Field(alias="yesterdayReward")
    """IDK. Clue exchange data."""
    y_crisis_ss: typing.Union[str, typing.Any] = pydantic.Field(alias="yCrisisSs", repr=False)
    """IDK."""
    medal_board: base.DDict = pydantic.Field(default_factory=base.DDict, alias="medalBoard")
    """Medal board."""


class ConsumableExpire(base.BaseModel):
    """Consumable expiration data."""

    ts: typing.Optional[base.ArknightsTimestamp]
    """When the consumable expires."""
    count: int
    """Amount of consumables."""


class User(base.BaseModel, extra="allow"):
    """User sync data. Not fully modeled."""

    status: Status
    """General user data."""
    troop: Troops
    """Operator data."""
    skin: Skins
    """Operator skin data."""
    social: Social
    """Data related to friends."""
    consumable: typing.Mapping[str, typing.Mapping[int, ConsumableExpire]] = {}
    """Consumable data."""
    inventory: typing.Mapping[str, int]
    """Inventory data. Item ID to amount.

    To access the static data for an item, use `client.gamedata.get_item(item_id)`.
    """



"""Arknights authentication client.

# Loading network configuration

As you likely know from the game itself, you need to get through two screens before entering the game.
The first checks the game version and network configuration.
The client and asset version must be sent along with a channel uid and access token to the game server.

| slug   | Example (en)                                                                    | Meaning                             |
| ------ | ------------------------------------------------------------------------------- | ----------------------------------- |
| gs     | https://gs.arknights.global:8443                                                | Game server                         |
| as     | https://as.arknights.global                                                     | Authentication server               |
| u8     | https://as.arknights.global/u8                                                  | u8* token authentication server     |
| hu     | https://ark-us-static-online.yo-star.com/assetbundle/official                   | Game assets                         |
| hv     | https://ark-us-static-online.yo-star.com/assetbundle/official/{0}/version       | Game assets & app versions          |
| rc     | https://ak-conf.arknights.global/config/prod/official/remote_config             | Unique server config                |
| an     | https://ark-us-static-online.yo-star.com/announce/{0}/announcement.meta.json    | Announcements                       |
| prean  | https://ark-us-static-online.yo-star.com/announce/{0}/preannouncement.meta.json | Pre-announcements on the login page |
| sl     | https://www.arknights.global/terms_of_service                                   | Terms of service                    |
| of     | https://www.arknights.global                                                    | Official webpage                    |
| pkgAd  | https://play.google.com/store/apps/details?id=com.YoStarEN.Arknights            | Google play store apk               |
| pkgIOS | https://apps.apple.com/us/app/id1464872022?mt=8                                 | IOS store apk                       |

*anyone know what u8 means?


# Authentication

Authentication allows you to get a session secret. Currently only email login is documented.

## Global (YoStar)

Global allows you to create guest accounts with only a device id.
The account however needs to bind a nickname before making any requests (after authentication).

### Get a permanent token
- Request a verification code to be sent to your email.
- Use the verification code to receive a yostar uid and token.
- Use the yostar uid and token to receive a channel uid and token

### Authenticate
- Use the channel uid and token to get an access token.
- Use the access token and channel uid to get an arknights player uid and u8 token.
- Use the arknights player uid and u8 token to get a session secret.


## CN (HyperGryph)

### Get a permanent token
- Send unhashed username and password to receive an access token.
- Optionally, send the access token to receive a channel uid.

### Authenticate
- Use the access token and channel uid to get an arknights player uid and u8 token.
- Use the arknights player uid and u8 token to get a session secret.


## Bilibili

### Get a permanent token
- Get a cipher key and hash from the bilibili api.
- Send a hashed password to receive a channel uid and access key.

### Authenticate
- Use the access key and channel uid to get an arknights player uid and u8 token.
- Use the arknights player uid and u8 token to get a session secret.


## Taiwan (Longcheng)

### Get a permanent token
Taiwan clearly has guest login and facebook / google play.
Unfortunately I have not been able to figure out how it works.

### Authenticate
- Use the channel uid and access token to get an arknights player uid and u8 token.
- Use the arknights player uid and u8 token to get a session secret.


# Making requests

Starting with getting the session secret, every request towards the game server needs to have:
- `secret` - session secret (except for the first request)
- `seqnum` - request sequence number (increments for every request, prevents concurrent requests)
- `uid` - arknights player uid

# Wrapper representation of sessions

The wrapper can represent a single or multiple sessions per client.
Single sessions are for getting personal data, but disallow concurrent requests.
Multiple sessions are for getting public data and allow concurrent requests.
"""
from __future__ import annotations

import abc
import asyncio
import base64
import dataclasses
import hashlib
import hmac
import json
import logging
import pathlib
import random
import string
import tempfile
import time
import typing
import urllib.parse
import uuid
import warnings

import aiohttp

from . import errors

__all__ = [
    "ArknightsDistributor",
    "ArknightsLanguage",
    "ArknightsServer",
    "Auth",
    "AuthSession",
    "BilibiliAuth",
    "CoreAuth",
    "GuestAuth",
    "HypergryphAuth",
    "MultiAuth",
    "NetworkSession",
    "YostarAuth",
]

logger: logging.Logger = logging.getLogger("arkprts.auth")

# these are in no way official slugs, just my own naming

ArknightsDistributor = typing.Literal["yostar", "hypergryph", "bilibili", "longcheng"]
ArknightsServer = typing.Literal["en", "jp", "kr", "cn", "bili", "tw"]
ArknightsLanguage = typing.Literal["en_US", "ja_JP", "ko_KR", "zh_CN", "zh_TW"]
ArknightsIdentifier = typing.Union[ArknightsDistributor, ArknightsServer, ArknightsLanguage]

ArknightsDomain = typing.Literal["gs", "as", "u8", "hu", "hv", "rc", "an", "prean", "sl", "of", "pkgAd", "pkgIOS"]
NETWORK_ROUTES: dict[ArknightsServer, str] = {
    "en": "https://ak-conf.arknights.global/config/prod/official/network_config",
    "jp": "https://ak-conf.arknights.jp/config/prod/official/network_config",
    "kr": "https://ak-conf.arknights.kr/config/prod/official/network_config",
    "cn": "https://ak-conf.hypergryph.com/config/prod/official/network_config",
    "bili": "https://ak-conf.hypergryph.com/config/prod/b/network_config",
    "tw": "https://ak-conf.txwy.tw/config/prod/official/network_config",
}
YOSTAR_PASSPORT_DOMAINS: dict[typing.Literal["en", "jp", "kr"], str] = {
    "en": "https://passport.arknights.global",
    "jp": "https://passport.arknights.jp",
    "kr": "https://passport.arknights.kr",
}

# the unity version is outdated, but it doesn't seem to matter
DEFAULT_HEADERS: typing.Mapping[str, str] = {
    "Content-Type": "application/json",
    "X-Unity-Version": "2017.4.39f1",
    "User-Agent": "Dalvik/2.1.0 (Linux; U; Android 11; KB2000 Build/RP1A.201005.001)",
    "Connection": "Keep-Alive",
}

# allows to user to pass in any bogus and still pass
REGION_IDENTIFIER_MAPPING: dict[
    ArknightsIdentifier,
    tuple[ArknightsDistributor, ArknightsServer, ArknightsLanguage] | None,
] = {
    # Distributor
    "yostar": ("yostar", "en", "en_US"),
    "hypergryph": ("hypergryph", "cn", "zh_CN"),
    "bilibili": ("bilibili", "bili", "zh_CN"),
    "longcheng": ("longcheng", "cn", "zh_TW"),
    # Server
    "en": ("yostar", "en", "en_US"),
    "jp": ("yostar", "jp", "ja_JP"),
    "kr": ("yostar", "kr", "ko_KR"),
    "cn": ("hypergryph", "cn", "zh_CN"),
    "bili": ("bilibili", "bili", "zh_CN"),
    "tw": ("longcheng", "cn", "zh_TW"),
    # Language
    "en_US": ("yostar", "en", "en_US"),
    "ja_JP": ("yostar", "jp", "ja_JP"),
    "ko_KR": ("yostar", "kr", "ko_KR"),
    "zh_CN": ("hypergryph", "cn", "zh_CN"),
    "zh_TW": ("longcheng", "cn", "zh_TW"),
}

RawAuthMapping = typing.TypedDict("RawAuthMapping", {"server": ArknightsServer, "channel_uid": str, "token": str})


def parse_server(identifier: ArknightsIdentifier) -> tuple[ArknightsDistributor, ArknightsServer, ArknightsLanguage]:
    """Parse a server, distributor, or language into a distributor, server, and language."""
    if identifier not in REGION_IDENTIFIER_MAPPING:
        raise ValueError(f"Invalid region, server, or language {identifier!r}")

    result = REGION_IDENTIFIER_MAPPING[identifier]
    if result is None:
        raise ValueError(f"{identifier} is not supported")

    return result


def create_random_device_ids() -> tuple[str, str, str]:
    """Create a random device id."""
    deviceid2 = "86" + "".join(random.choices(string.digits, k=13))
    return uuid.uuid4().hex, deviceid2, uuid.uuid4().hex


def generate_u8_sign(data: typing.Mapping[str, object]) -> str:
    """u8 auth sign."""
    query = urllib.parse.urlencode(sorted(data.items()))

    hama_code = hmac.new(b"91240f70c09a08a6bc72af1a5c8d4670", query.encode(), "sha1")
    return hama_code.hexdigest().lower()


# aiohttp uses a very noisy library
_charset_normalizer_logger = logging.getLogger("charset_normalizer")
_charset_normalizer_logger.setLevel(logging.INFO)


class NetworkSession:
    """Config-aware network session."""

    default_server: ArknightsServer | None = None
    """Default arknights server."""

    domains: dict[ArknightsServer, dict[ArknightsDomain, str]]
    """Arknights server domain routes."""
    versions: dict[ArknightsServer, dict[typing.Literal["resVersion", "clientVersion"], str]]
    """Arknights client versions."""

    def __init__(self, default_server: ArknightsServer | None = None) -> None:
        self.default_server = default_server

        self.domains = {server: {} for server in NETWORK_ROUTES}
        self.versions = {server: {} for server in NETWORK_ROUTES}

    async def raw_request(
        self,
        method: str,
        url: str,
        *,
        headers: typing.Mapping[str, str] | None = None,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Send a request to an arbitrary endpoint."""
        headers = {**DEFAULT_HEADERS, **(headers or {})}

        async with aiohttp.ClientSession() as session, session.request(method, url, headers=headers, **kwargs) as resp:
            try:
                data = await resp.json(content_type=None)
            except TypeError as e:
                resp.raise_for_status()
                raise errors.InvalidContentTypeError(await resp.text()) from e

            if data.get("error"):
                raise errors.GameServerError(data)

            if resp.status != 200:
                raise errors.InvalidStatusError(resp.status, data)

            return data

    async def request(
        self,
        domain: ArknightsDomain,
        endpoint: str | None = None,
        *,
        server: ArknightsServer | None = None,
        method: str | None = None,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Send a request to an arknights server."""
        server = server or self.default_server

        if "http" in domain:
            url = domain
        else:
            if server is None:
                raise ValueError("No default server set.")
            if server not in self.domains:
                raise ValueError(f"Invalid server {server!r}")
            if not self.domains[server]:
                await self.load_network_config(server)
            if domain not in self.domains[server]:
                raise ValueError(f"Invalid domain {domain!r}")

            url = self.domains[server][domain]

        if "{0}" in url:
            url = url.format("Android")  # iOS probably makes no difference
        if endpoint:
            url = url + "/" + endpoint

        if method is None:
            method = "POST" if kwargs.get("json") else "GET"

        data = await self.raw_request(method, url, **kwargs)

        if "result" in data and isinstance(data["result"], int) and data["result"] != 0:
            if "captcha" in data:
                raise errors.GeetestError(data)

            raise errors.ArkPrtsError(data)

        return data

    async def load_network_config(self, server: ArknightsServer | typing.Literal["all"] | None = None) -> None:
        """Load the network configuration."""
        server = server or self.default_server or "all"
        if server == "all":
            await asyncio.wait([asyncio.create_task(self.load_network_config(server)) for server in NETWORK_ROUTES])
            return

        logger.debug("Loading network configuration for %s.", server)
        data = await self.request(NETWORK_ROUTES[server])  # type: ignore # custom domain
        content = json.loads(data["content"])
        self.domains[server].update(content["configs"][content["funcVer"]]["network"])

    async def load_version_config(self, server: ArknightsServer | typing.Literal["all"] | None = None) -> None:
        """Load the version configuration."""
        server = server or self.default_server or "all"
        if server == "all":
            await asyncio.wait([asyncio.create_task(self.load_version_config(server)) for server in NETWORK_ROUTES])
            return

        logger.debug("Loading version configuration for %s.", server)
        data = await self.request("hv", server=server)
        self.versions[server].update(data)


@dataclasses.dataclass()
class AuthSession:
    """An already authenticated session."""

    server: ArknightsServer
    """Arknights server."""
    uid: str
    """Arknights user UID."""
    secret: str = ""
    """Arknights session token."""
    seqnum: int = 1
    """Request sequence number."""
    lock: asyncio.Lock = dataclasses.field(default_factory=asyncio.Lock)
    """Lock to prevent concurrent requests."""

    @property
    def is_locked(self) -> bool:
        """Whether the session is currently making a request."""
        return self.lock.locked()

    async def __aenter__(self) -> typing.Mapping[str, str]:
        """Prepare for the next request and return headers."""
        await self.lock.acquire()
        self.seqnum += 1
        return {"secret": self.secret, "seqnum": str(self.seqnum), "uid": self.uid}

    async def __aexit__(self, *exc: object) -> None:
        """Release lock."""
        self.lock.release()


class CoreAuth(typing.Protocol):
    """Authentication client typing protocol.

    Look for subclasses for more specific authentication methods.
    """

    network: NetworkSession
    """Network session."""

    async def auth_request(
        self,
        endpoint: str,
        *,
        region: ArknightsServer | None = None,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Send an authenticated request to the arkights game server."""


class Auth(abc.ABC, CoreAuth):
    """Authentication client for single sessions."""

    server: ArknightsServer
    """Arknights region."""
    distributor: ArknightsDistributor
    """Arknights server."""
    network: NetworkSession
    """Network session."""
    device_ids: tuple[str, str, str]
    """Device ids."""
    session: AuthSession
    """Authentication session."""

    def __init__(
        self,
        server: ArknightsServer | None = None,
        *,
        network: NetworkSession | None = None,
    ) -> None:
        if server is None and network is not None:
            server = network.default_server

        self.distributor, self.server, _ = parse_server(server or "en")
        self.network = network or NetworkSession(default_server=self.server)
        self.session = AuthSession(self.server, "", "")
        self.device_ids = create_random_device_ids()

    @property
    def uid(self) -> str:
        """Arknights user UID."""
        return self.session.uid

    @property
    def secret(self) -> str:
        """Arknights session token."""
        return self.session.secret

    async def request(
        self,
        domain: ArknightsDomain,
        endpoint: str | None = None,
        *,
        server: ArknightsServer | None = None,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Send a request to an arknights server."""
        if server and server != self.server:
            raise ValueError(f"Single-session client is bound to {self.server!r} server.")

        return await self.network.request(domain, endpoint, server=self.server, **kwargs)

    async def auth_request(
        self,
        endpoint: str,
        *,
        server: ArknightsServer | None = None,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Send an authenticated request to the arknights game server."""
        if server and server != self.server:
            raise ValueError(f"Single-session client is bound to {self.server!r} region.")

        if not self.session.uid:
            raise errors.NotLoggedInError("Not logged in.")

        async with self.session as headers:
            logger.debug("[UID: %s] Sending request #%s to %s.", self.uid, headers["seqnum"], endpoint)
            return await self.request("gs", endpoint, headers=headers, server=self.server, **kwargs)

    async def _get_u8_token(
        self,
        channel_uid: str,
        access_token: str,
    ) -> tuple[str, str]:
        """Get an arknights uid and u8 token from a channel uid and access token."""
        logger.debug("Getting u8 token for %s.", channel_uid)
        channel_id = {"hypergryph": "1", "bilibili": "2", "yostar": "3"}[self.distributor]
        if self.distributor == "yostar":
            extension = {"uid": channel_uid, "token": access_token}
        else:
            extension = {"uid": channel_uid, "access_token": access_token}

        body = {
            "appId": "1",
            "platform": 1,
            "channelId": channel_id,
            "subChannel": channel_id,
            "extension": json.dumps(extension),
            # optional fields:
            "worldId": channel_id,
            "deviceId": self.device_ids[0],
            "deviceId2": self.device_ids[1],
            "deviceId3": self.device_ids[2],
        }
        # optional:
        body["sign"] = generate_u8_sign(body)

        data = await self.request("u8", "user/v1/getToken", json=body)
        uid, token = data["uid"], data["token"]
        self.session.uid = uid
        return uid, token

    async def _get_secret(
        self,
        uid: str,
        u8_token: str,
    ) -> str:
        """Get a secret from an arknights uid and a u8 token."""
        logger.debug("Getting session secret for %s.", uid)
        if not self.network.versions.get(self.server):
            await self.network.load_version_config(self.server)

        network_version = {"hypergryph": "5", "bilibili": "5", "yostar": "1"}[self.distributor]

        body = {
            "platform": 1,
            "networkVersion": network_version,
            "assetsVersion": self.network.versions[self.server]["resVersion"],
            "clientVersion": self.network.versions[self.server]["clientVersion"],
            "token": u8_token,
            "uid": uid,
            "deviceId": self.device_ids[0],
            "deviceId2": self.device_ids[1],
            "deviceId3": self.device_ids[2],
        }
        headers = {
            "secret": "",
            "seqnum": "1",
            "uid": uid,
        }
        data = await self.request("gs", "account/login", json=body, headers=headers)
        secret = data["secret"]
        self.session.secret = secret
        logger.info("Logged in with UID %s", uid)
        return secret

    @classmethod
    async def from_token(
        cls,
        region: ArknightsServer,
        channel_uid: str,
        token: str,
        *,
        network: NetworkSession | None = None,
    ) -> Auth:
        """Create a client from a token."""
        if region in ("en", "jp", "kr"):
            auth = YostarAuth(region, network=network)
            await auth.login_with_token(channel_uid, token)
        elif region == "cn":
            auth = HypergryphAuth(region, network=network)
            await auth.login_with_token(channel_uid, token)
        elif region == "bili":
            auth = BilibiliAuth(region, network=network)
            await auth.login_with_token(channel_uid, token)
        elif region == "tw":
            auth = LongchengAuth(region, network=network)
            await auth.login_with_token(channel_uid, token)
        else:
            raise ValueError(f"Cannot create a generic auth client for region {region!r}")

        return auth


class YostarAuth(Auth):
    """Authentication client for global accounts."""

    distributor: typing.Literal["yostar"]
    server: typing.Literal["en", "jp", "kr"]

    def __init__(
        self,
        server: typing.Literal["en", "jp", "kr"] = "en",
        *,
        network: NetworkSession | None = None,
    ) -> None:
        super().__init__(server, network=network)

    async def request_passport(self, endpoint: str, **kwargs: typing.Any) -> typing.Any:
        """Send a request to a yostar passport endpoint."""
        return await self.request(YOSTAR_PASSPORT_DOMAINS[self.server], endpoint, **kwargs)  # type: ignore  # custom domain

    async def _get_access_token(self, channel_uid: str, yostar_token: str) -> str:
        """Get an access token from a channel uid and yostar token."""
        body = {
            "platform": "android",
            "uid": channel_uid,
            "token": yostar_token,
            "deviceId": self.device_ids[0],
        }
        data = await self.request_passport("user/login", json=body)
        return data["accessToken"]

    async def _request_yostar_auth(self, email: str) -> None:
        """Request to log in with a yostar account."""
        logger.debug("Sending code to %s.", email)
        body = {"platform": "android", "account": email, "authlang": "en"}
        await self.request_passport("account/yostar_auth_request", json=body)

    async def _submit_yostar_auth(self, email: str, code: str) -> tuple[str, str]:
        """Submit a yostar auth code and receieve a yostar uid and yostar token."""
        body = {"account": email, "code": code}
        data = await self.request_passport("account/yostar_auth_submit", json=body)
        return data["yostar_uid"], data["yostar_token"]

    async def _get_yostar_token(self, email: str, yostar_uid: str, yostar_token: str) -> tuple[str, str]:
        """Get a channel uid and yostar token from a yostar uid and yostar token."""
        body = {
            "yostar_username": email,
            "yostar_uid": yostar_uid,
            "yostar_token": yostar_token,
            "deviceId": self.device_ids[0],
            "createNew": "0",
        }
        data = await self.request_passport("user/yostar_createlogin", json=body)
        return data["uid"], data["token"]

    async def create_guest_account(self) -> tuple[str, str]:
        """Create a new guest account."""
        body = {
            "deviceId": self.device_ids[0],
        }
        data = await self.request_passport("user/create", json=body)
        logger.debug("Created guest account %s", data["uid"])
        return data["uid"], data["token"]

    async def _bind_nickname(self, nickname: str) -> None:
        """Bind a nickname. Required for new accounts."""
        logger.debug("Binding nickname of %s to %r.", self.uid, nickname)
        await self.auth_request("user/bindNickName", json={"nickName": nickname})

    async def login_with_token(self, channel_uid: str, yostar_token: str) -> None:
        """Login with a yostar token."""
        access_token = await self._get_access_token(channel_uid, yostar_token)
        self.session.uid, u8_token = await self._get_u8_token(channel_uid, access_token)
        await self._get_secret(self.session.uid, u8_token)

    async def get_token_from_email_code(
        self,
        email: str | None = None,
        code: str | None = None,
        *,
        stdin: bool = False,
    ) -> tuple[str, str]:
        """Get a token from a yostar account."""
        if not email:
            if not stdin:
                raise TypeError("Email not provided but stdin is disabled.")

            email = input("Enter email:")

        if not code:
            await self._request_yostar_auth(email)
            if not stdin:
                return "", ""

            print(f"Code sent to {email}.")  # noqa: T201
            code = input("Enter code: ")

        yostar_uid, yostar_token = await self._submit_yostar_auth(email, code)
        return await self._get_yostar_token(email, yostar_uid, yostar_token)

    async def login_with_email_code(
        self,
        email: str | None = None,
        code: str | None = None,
        *,
        stdin: bool = True,
    ) -> tuple[str, str]:
        """Login with a yostar account. Uses stdin by default."""
        channel_uid, token = await self.get_token_from_email_code(email, code, stdin=stdin)
        await self.login_with_token(channel_uid, token)

        if stdin:
            print(f"Channel UID: {channel_uid} Token: {token}")  # noqa: T201
            print(f'Usage: login_with_token("{channel_uid}", "{token}")')  # noqa: T201

        return channel_uid, token

    async def login_as_guest(self, nickname: str | None = None) -> tuple[str, str]:
        """Login as guest and return tokens."""
        channel_uid, yostar_token = await self.create_guest_account()
        await self.login_with_token(channel_uid, yostar_token)
        await self._bind_nickname(nickname or "Doctor")
        return channel_uid, yostar_token


class HypergryphAuth(Auth):
    """Authentication client for chinese accounts."""

    distributor: typing.Literal["hypergryph"]
    server: typing.Literal["cn"]

    def __init__(
        self,
        server: typing.Literal["cn"] = "cn",
        *,
        network: NetworkSession | None = None,
    ) -> None:
        super().__init__(server, network=network)

    async def _get_hypergryph_access_token(self, username: str, password: str) -> str:
        """Get an access token from a username and password."""
        data = {
            "account": username,
            "password": password,
            "deviceId": self.device_ids[0],
            "platform": 1,
        }
        data["sign"] = generate_u8_sign(data)
        data = await self.network.request("as", "user/login", json=data)
        return data["token"]

    async def _get_hypergryph_uid(self, token: str) -> str:
        """Get a channel uid from a hypergryph access token."""
        data = {"token": token}
        data["sign"] = generate_u8_sign(data)
        data = await self.network.request("as", "user/auth", json=data)
        return data["uid"]

    async def login_with_token(self, channel_uid: str, access_token: str) -> None:
        """Login with an access token."""
        self.session.uid, u8_token = await self._get_u8_token(channel_uid, access_token)
        await self._get_secret(self.session.uid, u8_token)

    async def get_token_from_password(
        self,
        username: str | None = None,
        password: str | None = None,
        *,
        stdin: bool = False,
    ) -> tuple[str, str]:
        """Get a token from a hypergryph account."""
        if not username or not password:
            if not stdin:
                raise TypeError("Password not provided but stdin is disabled.")

            username = input("Enter username: ")
            password = input("Enter password: ")

        access_token = await self._get_hypergryph_access_token(username, password)
        channel_uid = await self._get_hypergryph_uid(access_token)
        return channel_uid, access_token

    async def login(
        self,
        username: str | None = None,
        password: str | None = None,
        *,
        stdin: bool = True,
    ) -> tuple[str, str]:
        """Login with a hypergryph account."""
        channel_uid, access_token = await self.get_token_from_password(username, password, stdin=stdin)
        await self.login_with_token(channel_uid, access_token)

        if stdin:
            print(f"Channel UID: {channel_uid} Access token: {access_token}")  # noqa: T201
            print(f'Usage: login_with_token("{channel_uid}", "{access_token}")')  # noqa: T201

        return channel_uid, access_token


class BilibiliAuth(Auth):
    """Authentication client for bilibili accounts."""

    distributor: typing.Literal["bilibili"]
    server: typing.Literal["bili"]

    cipher_key: str
    """Bilibili pkcs1 openssl key for arknights."""
    password_hash: str
    """Extra text to go at the start of the password."""

    def __init__(
        self,
        server: typing.Literal["bili"] = "bili",
        *,
        network: NetworkSession | None = None,
    ) -> None:
        super().__init__(server, network=network)

    @staticmethod
    def _sign_body(body: typing.Mapping[str, str]) -> str:
        """Sign request body."""
        body = dict(sorted(body.items()))
        string = "".join(body.values()) + "8783abfb533544c59e598cddc933d1bf"
        return hashlib.md5(string.encode()).hexdigest()

    def _sign_password(self, password: str) -> str:
        """Sign password."""
        import rsa

        public_key = rsa.PublicKey.load_pkcs1_openssl_pem(self.cipher_key.encode())
        signed = rsa.encrypt((self.password_hash + password).encode(), public_key)
        return base64.b64encode(signed).decode()

    async def _load_cipher(self) -> None:
        """Load the cipher key and hash."""
        body = dict(
            merchant_id="328",
            game_id="952",
            server_id="1178",
            version="3",
            timestamp=str(int(time.time())),
            cipher_type="bili_login_rsa",
        )
        body["sign"] = self._sign_body(body)
        async with aiohttp.request(
            "POST",
            "https://line1-sdk-center-login-sh.biligame.net/api/external/issue/cipher/v3",
            data=body,
        ) as r:
            data = await r.json()

        self.cipher_key, self.password_hash = data["cipher_key"], data["hash"]

    async def _get_access_key(self, username: str, password: str, *, bd_id: str | None = None) -> tuple[str, str]:
        """Get an access key from username and password."""
        await self._load_cipher()

        bd_id = bd_id or "-".join(random.randbytes(round(n / 2)).hex()[:n] for n in (8, 4, 4, 4, 12, 8, 4, 4, 4, 3))
        body = dict(
            merchant_id="328",
            game_id="952",
            server_id="1178",
            version="3",
            timestamp=str(int(time.time())),
            bd_id=bd_id,
            user_id=username,
            pwd=self._sign_password(password),
        )
        body["sign"] = self._sign_body(body)
        async with aiohttp.request(
            "POST",
            "https://line1-sdk-center-login-sh.biligame.net/api/external/login/v3",
            data=body,
        ) as r:
            data = await r.json()

        return data["uid"], data["access_key"]

    async def login_with_token(self, channel_uid: str, access_token: str) -> None:
        """Login with an access key."""
        self.session.uid, u8_token = await self._get_u8_token(channel_uid, access_token)
        await self._get_secret(self.session.uid, u8_token)

    async def get_token_from_password(
        self,
        username: str | None = None,
        password: str | None = None,
        bd_id: str | None = None,
        *,
        stdin: bool = False,
    ) -> tuple[str, str]:
        """Get a token from a bilibili account."""
        if not username or not password:
            if not stdin:
                raise TypeError("Password not provided but stdin is disabled.")

            username = input("Enter username: ")
            password = input("Enter password: ")

        return await self._get_access_key(username, password, bd_id=bd_id)

    async def login(
        self,
        username: str | None = None,
        password: str | None = None,
        bd_id: str | None = None,
        *,
        stdin: bool = True,
    ) -> tuple[str, str]:
        """Login with a bilibili account."""
        channel_uid, access_token = await self.get_token_from_password(username, password, bd_id, stdin=stdin)
        await self.login_with_token(channel_uid, access_token)

        if stdin:
            print(f"Channel UID: {channel_uid} Access key: {access_token}")  # noqa: T201
            print(f'Usage: login_with_token("{channel_uid}", "{access_token}")')  # noqa: T201

        return channel_uid, access_token


class LongchengAuth(Auth):
    """Authentication client for taiwan accounts."""

    distributor: typing.Literal["longcheng"]
    server: typing.Literal["tw"]

    def __init__(
        self,
        server: typing.Literal["tw"] = "tw",
        *,
        network: NetworkSession | None = None,
    ) -> None:
        super().__init__(server, network=network)

    async def login_with_token(self, channel_uid: str, access_token: str) -> None:
        """Login with an access token."""
        self.session.uid, u8_token = await self._get_u8_token(channel_uid, access_token)
        await self._get_secret(self.session.uid, u8_token)


class MultiAuth(CoreAuth):
    """Authentication client for multiple sessions."""

    network: NetworkSession
    """Network session."""

    # may be exceeded if multiple sessions are created at once
    max_sessions: int
    """Maximum number of concurrent sessions per region."""
    sessions: list[AuthSession]
    """Authentication sessions."""

    def __init__(
        self,
        max_sessions: int = 6,
        *,
        network: NetworkSession | None = None,
    ) -> None:
        self.network = network or NetworkSession()
        self.max_sessions = max_sessions
        self.sessions = []

    def _get_free_session(self, server: ArknightsServer) -> AuthSession | None:
        """Get a free session in a server."""
        for session in self.sessions:
            if session.server == server and not session.is_locked:
                return session

        return None

    async def _wait_for_free_session(self, server: ArknightsServer) -> AuthSession:
        """Wait a session to be freed."""
        while True:
            await asyncio.sleep(0.1)
            if session := self._get_free_session(server):
                return session

    async def _create_new_session(self, server: ArknightsServer) -> AuthSession:
        """Create a new session for a selected server."""
        raise RuntimeError("No method for creating new sessions specified.")

    async def request(
        self,
        domain: ArknightsDomain,
        endpoint: str | None = None,
        *,
        server: ArknightsServer | None = None,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Send a request to an arknights server."""
        return await self.network.request(domain, endpoint, server=server, **kwargs)

    async def auth_request(
        self,
        endpoint: str,
        *,
        server: ArknightsServer | None = None,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Send an authenticated request to the arknights game server."""
        server = server or self.network.default_server
        if server is None:
            raise ValueError("No default server set.")

        session = self._get_free_session(server)
        if session is None and sum(session.server == server for session in self.sessions) >= self.max_sessions:
            session = await self._wait_for_free_session(server)
        if session is None:
            session = await self._create_new_session(server)
            self.sessions.append(session)
            logger.debug("Created new session %s for server %s.", session.uid, server)

        async with session as headers:
            logger.debug(
                "[GUEST UID: %s %s] Sending request #%s to %s.",
                session.uid,
                server,
                headers["seqnum"],
                endpoint,
            )
            return await self.request("gs", endpoint, headers=headers, server=server, **kwargs)

    def add_session(self, session: AuthSession | Auth | MultiAuth | None) -> None:
        """Add a session to the list of sessions."""
        if isinstance(session, AuthSession):
            self.sessions.append(session)
        elif isinstance(session, Auth):
            self.sessions.append(session.session)
        elif isinstance(session, MultiAuth):
            self.sessions.extend(session.sessions)
        else:
            raise TypeError(f"Invalid session type {type(session)}")


class GuestAuth(MultiAuth):
    """Authentication client for dynamically generating guest accounts."""

    cache_path: pathlib.Path | None
    """Location of stored guest authentication."""
    upcoming_auth: list[RawAuthMapping]
    """Upcoming accounts that are yet to be loaded."""

    def __init__(
        self,
        max_sessions: int = 6,
        cache: pathlib.Path | str | typing.Sequence[RawAuthMapping] | typing.Literal[False] | None = None,
        *,
        network: NetworkSession | None = None,
    ) -> None:
        super().__init__(max_sessions=max_sessions, network=network)

        # load cache file or use provided auth
        self.upcoming_auth = []
        if cache is False:
            self.cache_path = None
        elif isinstance(cache, (pathlib.Path, str)):
            self.cache_path = pathlib.Path(cache)
        elif cache is None:
            self.cache_path = pathlib.Path(tempfile.gettempdir()) / "arkprts_auth_cache.json"
        else:
            self.cache_path = None
            self.upcoming_auth = list(cache)

        if self.cache_path:
            self.upcoming_auth.extend(self._load_cache())

    def _load_cache(self) -> typing.Sequence[RawAuthMapping]:
        """Load cached guest accounts."""
        if not self.cache_path:
            return []

        if not self.cache_path.exists():
            return []

        with self.cache_path.open() as f:
            data = json.load(f)

        return data

    def _save_cache(self, data: typing.Sequence[RawAuthMapping]) -> None:
        """Save cached guest accounts."""
        if not self.cache_path:
            return

        with self.cache_path.open("w") as f:
            json.dump(data, f)

    def _append_to_cache(self, server: ArknightsServer, channel_uid: str, token: str) -> None:
        """Append a guest account to the cache."""
        if not self.cache_path:
            return

        data = list(self._load_cache())
        data.append({"server": server, "channel_uid": channel_uid, "token": token})
        self._save_cache(data)

    async def _load_upcoming_session(self, server: ArknightsServer) -> AuthSession | None:
        """Take one upcoming auth and create a session from it."""
        for i, auth in enumerate(self.upcoming_auth):
            if auth["server"] == server:
                self.upcoming_auth.pop(i)
                break
        else:
            return None

        logging.debug("Loading cached auth %s for %s.", auth["channel_uid"], auth["server"])
        try:
            auth = await Auth.from_token(server, auth["channel_uid"], auth["token"], network=self.network)
        except errors.ArkPrtsError as e:
            warnings.warn(f"Failed to load cached auth: {e}")
            # remove faulty auth from cache file
            data = list(self._load_cache())
            data.remove(auth)
            self._save_cache(data)

            return None

        return auth.session

    async def _create_new_session(self, server: ArknightsServer) -> AuthSession:
        """Create a new guest account."""
        if server not in ("en", "jp", "kr"):
            raise ValueError("Guest accounts are only supported on the global server.")

        session = await self._load_upcoming_session(server)
        if session is not None:
            return session

        auth = YostarAuth(server, network=self.network)
        channel_uid, token = await auth.login_as_guest()
        self._append_to_cache(server=server, channel_uid=channel_uid, token=token)
        return auth.session



"""Test authentication."""
import asyncio
import collections
import os
import typing
import warnings

import pytest

import arkprts


@pytest.fixture(scope="session")
async def network() -> arkprts.NetworkSession:
    """Network session."""
    return arkprts.NetworkSession()


async def test_config(network: arkprts.NetworkSession) -> None:
    await network.load_version_config("all")

    assert all(versions for versions in network.versions.values())
    assert all(domains for domains in network.domains.values())


@pytest.fixture(scope="session")
async def client(network: arkprts.NetworkSession) -> arkprts.Client:
    """Public global client."""
    return arkprts.Client(arkprts.GuestAuth(max_sessions=1, cache=os.environ.get("GUEST_AUTH_CACHE"), network=network))


@pytest.fixture(scope="session")
async def en_client(pytestconfig: pytest.Config, network: arkprts.NetworkSession) -> arkprts.Client:
    """Private en client."""
    channel_uid, yostar_token = os.environ.get("EN_CHANNEL_UID"), os.environ.get("EN_YOSTAR_TOKEN")
    if not channel_uid or not yostar_token:
        if pytestconfig.getoption("--runblocking") and os.environ.get("EN_EMAIL"):
            auth = arkprts.YostarAuth("en", network=network)
            channel_uid, yostar_token = await auth.login_with_email_code(os.environ["EN_EMAIL"])
            warnings.warn(f"Please use EN_CHANNEL_UID={channel_uid} EN_YOSTAR_TOKEN={yostar_token}")
            return arkprts.Client(auth)

        pytest.skip("EN_CHANNEL_UID or EN_YOSTAR_TOKEN not set (use --runblocking if you want to use email)")

    auth = await arkprts.Auth.from_token("en", channel_uid, yostar_token, network=network)
    return arkprts.Client(auth)


@pytest.fixture(scope="session")
async def jp_client(pytestconfig: pytest.Config, network: arkprts.NetworkSession) -> arkprts.Client:
    """Private jp client."""
    channel_uid, yostar_token = os.environ.get("JP_CHANNEL_UID"), os.environ.get("JP_YOSTAR_TOKEN")
    if not channel_uid or not yostar_token:
        if pytestconfig.getoption("--runblocking") and os.environ.get("JP_EMAIL"):
            auth = arkprts.YostarAuth("jp", network=network)
            channel_uid, yostar_token = await auth.login_with_email_code(os.environ["JP_EMAIL"])
            warnings.warn(f"Please use JP_CHANNEL_UID={channel_uid} JP_YOSTAR_TOKEN={yostar_token}")
            return arkprts.Client(auth)

        pytest.skip("JP_CHANNEL_UID or JP_YOSTAR_TOKEN not set")

    auth = await arkprts.Auth.from_token("jp", channel_uid, yostar_token, network=network)
    return arkprts.Client(auth)


@pytest.fixture(scope="session")
async def kr_client(pytestconfig: pytest.Config, network: arkprts.NetworkSession) -> arkprts.Client:
    """Private kr client."""
    channel_uid, yostar_token = os.environ.get("KR_CHANNEL_UID"), os.environ.get("KR_YOSTAR_TOKEN")
    if not channel_uid or not yostar_token:
        if pytestconfig.getoption("--runblocking") and os.environ.get("KR_EMAIL"):
            auth = arkprts.YostarAuth("kr", network=network)
            channel_uid, yostar_token = await auth.login_with_email_code(os.environ["KR_EMAIL"])
            warnings.warn(f"Please use KR_CHANNEL_UID={channel_uid} KR_YOSTAR_TOKEN={yostar_token}")
            return arkprts.Client(auth)

        pytest.skip("KR_CHANNEL_UID or KR_YOSTAR_TOKEN not set")

    auth = await arkprts.Auth.from_token("kr", channel_uid, yostar_token, network=network)
    return arkprts.Client(auth)


@pytest.fixture(scope="session")
async def cn_client(network: arkprts.NetworkSession) -> arkprts.Client:
    """Private cn client."""
    channel_uid, access_token = os.environ.get("CN_CHANNEL_UID"), os.environ.get("CN_ACCESS_TOKEN")
    if not channel_uid or not access_token:
        cn_username, cn_password = os.environ.get("CN_USERNAME"), os.environ.get("CN_PASSWORD")
        if cn_username:
            auth = arkprts.HypergryphAuth(network=network)
            channel_uid, access_token = await auth.login(cn_username, cn_password)
            warnings.warn(f"Please use CN_CHANNEL_UID={channel_uid} CN_ACCESS_TOKEN={access_token}")
            return arkprts.Client(auth)

        pytest.skip("CN_CHANNEL_UID or CN_ACCESS_TOKEN not set")

    auth = await arkprts.Auth.from_token("cn", channel_uid, access_token, network=network)
    return arkprts.Client(auth)


@pytest.fixture(scope="session")
async def bili_client(network: arkprts.NetworkSession) -> arkprts.Client:
    """Private bili client."""
    channel_uid, access_token = os.environ.get("BILI_CHANNEL_UID"), os.environ.get("BILI_ACCESS_TOKEN")
    if not channel_uid or not access_token:
        bili_username, bili_password = os.environ.get("BILI_USERNAME"), os.environ.get("BILI_PASSWORD")
        bili_bd_id = os.environ.get("BILI_BD_ID")
        if bili_username:
            auth = arkprts.BilibiliAuth(network=network)
            channel_uid, access_token = await auth.login(bili_username, bili_password, bd_id=bili_bd_id)
            warnings.warn(f"Please use BILI_CHANNEL_UID={channel_uid} BILI_ACCESS_TOKEN={access_token}")
            return arkprts.Client(auth)

        pytest.skip("BILI_CHANNEL_UID or BILI_ACCESS_TOKEN not set")

    auth = await arkprts.Auth.from_token("bili", channel_uid, access_token, network=network)
    return arkprts.Client(auth)


@pytest.fixture(scope="session")
async def tw_client(network: arkprts.NetworkSession) -> arkprts.Client:
    """Private tw client."""
    channel_uid, access_token = os.environ.get("TW_CHANNEL_UID"), os.environ.get("TW_ACCESS_TOKEN")
    if not channel_uid or not access_token:
        pytest.skip("TW_CHANNEL_UID or TW_ACCESS_TOKEN not set")

    auth = await arkprts.Auth.from_token("tw", channel_uid, access_token, network=network)
    return arkprts.Client(auth)


def test_jp_client(jp_client: arkprts.Client) -> None:
    assert isinstance(jp_client.auth, arkprts.Auth)
    assert jp_client.auth.server == "jp"
    assert jp_client.auth.distributor == "yostar"
    assert jp_client.auth.session.uid


def test_kr_client(kr_client: arkprts.Client) -> None:
    assert isinstance(kr_client.auth, arkprts.Auth)
    assert kr_client.auth.server == "kr"
    assert kr_client.auth.distributor == "yostar"
    assert kr_client.auth.session.uid


def test_cn_client(cn_client: arkprts.Client) -> None:
    assert isinstance(cn_client.auth, arkprts.Auth)
    assert cn_client.auth.server == "cn"
    assert cn_client.auth.distributor == "hypergryph"
    assert cn_client.auth.session.uid


def test_bili_client(bili_client: arkprts.Client) -> None:
    assert isinstance(bili_client.auth, arkprts.Auth)
    assert bili_client.auth.server == "bili"
    assert bili_client.auth.distributor == "bilibili"
    assert bili_client.auth.session.uid


def test_tw_client(tw_client: arkprts.Client) -> None:
    assert isinstance(tw_client.auth, arkprts.Auth)
    assert tw_client.auth.server == "tw"
    assert tw_client.auth.distributor == "longcheng"
    assert tw_client.auth.session.uid


class MockGuestAuth(arkprts.GuestAuth):
    async def request(
        self,
        domain: arkprts.auth.ArknightsDomain,
        endpoint: typing.Optional[str] = None,
        **kwargs: typing.Any,
    ) -> typing.Any:
        if endpoint == "...":
            return await super().request("rc", **kwargs)

        return await super().request(domain, endpoint, **kwargs)


async def test_guest_auth(network: arkprts.NetworkSession) -> None:
    auth = MockGuestAuth(max_sessions=2, cache=os.environ.get("GUEST_AUTH_CACHE"), network=network)

    def count() -> collections.Counter[str]:
        return collections.Counter(session.server for session in auth.sessions)

    assert len(auth.sessions) == 0

    await auth.auth_request("...", server="en")
    assert len(auth.sessions) == 1
    assert count()["en"] == 1

    await auth.auth_request("...", server="jp")
    assert len(auth.sessions) == 2
    assert count()["jp"] == 1

    await auth.auth_request("...", server="en")
    assert len(auth.sessions) == 2
    assert count()["en"] == 1

    await asyncio.gather(
        auth.auth_request("...", server="en"),
        auth.auth_request("...", server="en"),
    )
    assert len(auth.sessions) == 3
    assert count()["en"] == 2



"""Automation client. Potentially bannable."""
import base64
import binascii
import hashlib
import io
import json
import random
import string
import time
import typing
import zipfile

from arkprts import auth as authn
from arkprts import gamedata as gd

from .client import CoreClient

__all__ = ["AutomationClient"]


def recursively_update_dict(
    target: typing.MutableMapping[typing.Any, typing.Any],
    source: typing.Mapping[typing.Any, typing.Any],
) -> None:
    """Recursively update a dictionary.

    This is used to update the player data.
    """
    for key, value in source.items():
        if isinstance(value, dict):
            recursively_update_dict(target[key], typing.cast("dict[object, object]", value))
        elif isinstance(value, list):
            for i, v in enumerate(typing.cast("list[object]", value)):
                if isinstance(v, dict):
                    recursively_update_dict(target[key][i], typing.cast("dict[object, object]", v))
        else:
            target[key] = value


def recursively_delete_dict(
    target: typing.MutableMapping[typing.Any, typing.Any],
    source: typing.Mapping[typing.Any, typing.Any],
) -> None:
    """Recursively delete items from a dictionary.

    This is used to update the player data.
    """
    for key, value in source.items():
        if isinstance(value, dict):
            recursively_delete_dict(target[key], typing.cast("dict[object, object]", value))
        elif isinstance(value, list):
            for i, v in enumerate(typing.cast("list[object]", value)):
                if isinstance(v, dict):
                    recursively_delete_dict(target[key][i], typing.cast("dict[object, object]", v))
        else:
            del target[key]


# https://github.com/Rhine-Department-0xf/Rhine-DFramwork/blob/main/client/encryption.py


def get_md5(data: str | bytes) -> str:
    """Get the MD5 hash of a string or bytes object."""
    if isinstance(data, str):
        data = data.encode()

    return hashlib.md5(data).hexdigest()


def rijndael_encrypt(data: bytes, key: bytes, iv: bytes) -> bytes:
    from Crypto.Cipher import AES
    from Crypto.Util.Padding import pad

    aes_obj = AES.new(key, AES.MODE_CBC, iv)  # pyright: ignore[reportUnknownMemberType]
    encrypt_buf = aes_obj.encrypt(pad(data, AES.block_size))
    return encrypt_buf


def rijndael_decrypt(data: bytes, key: bytes, iv: bytes) -> bytes:
    from Crypto.Cipher import AES
    from Crypto.Util.Padding import unpad

    aes_obj = AES.new(key, AES.MODE_CBC, iv)  # pyright: ignore[reportUnknownMemberType]
    decrypt_buf = aes_obj.decrypt(data)
    return unpad(decrypt_buf, AES.block_size)


def encrypt_battle_data(data: str, login_time: int) -> str:
    iv = "".join(random.choices(string.ascii_letters + string.digits, k=16)).encode()
    key_array = bytearray.fromhex(get_md5(f"pM6Umv*^hVQuB6t&{login_time}"))
    return binascii.hexlify(rijndael_encrypt(data.encode(), key_array, iv) + iv).decode().upper()


def decrypt_battle_data(data: str, login_time: int) -> typing.Any:
    battle_data = data[:-32:]
    battle_data_array = bytearray.fromhex(battle_data)
    iv = data[-32::]
    iv_array = bytearray.fromhex(iv)
    key_array = bytearray.fromhex(get_md5(f"pM6Umv*^hVQuB6t&{login_time}"))
    decrypted_data = rijndael_decrypt(battle_data_array, key_array, iv_array)
    return json.loads(decrypted_data)


def decrypt_battle_replay(battle_replay: str) -> typing.Any:
    """Decrypt a battle replay to get battle_id.

    op: 0 = place, 1 = retreat, 2 = skill
    direction: 0 = up, 1 = right 2 = down, 3 = left
    pos: 1,1 = bottom left (cartesian)
    """
    data = base64.b64decode(battle_replay)
    with zipfile.ZipFile(io.BytesIO(data), "r") as z, z.open("default_entry") as f:
        return json.load(f)


def get_battle_data_access(login_time: int, hash_key: str) -> str:
    """Get battle data access for battle finish stats."""
    # I have no idea what this is for actually
    return get_md5(f"{hash_key}{login_time}").upper()


def encrypt_battle_id(battle_id: str) -> str:
    """Encrypt a battle ID to get is_cheat."""
    data = bytearray(i + 7 for i in battle_id.encode())
    return base64.b64encode(data).decode()


class AutomationClient(CoreClient):
    """Automation client. Potentially bannable.

    Does not use any models.
    """

    data: typing.Any
    """Player data."""
    time_offset: int
    """Time offset in seconds."""

    def __init__(
        self,
        auth: authn.Auth | None = None,
        *,
        gamedata: gd.GameData | None = None,
        network: authn.NetworkSession | None = None,
        server: authn.ArknightsServer | None = None,
        language: authn.ArknightsLanguage | None = None,
    ) -> None:
        super().__init__(auth, gamedata=gamedata, network=network, server=server, language=language)
        self.data = {}

    def update_player_data_delta(
        self,
        delta: typing.Mapping[typing.Literal["deleted", "modified"], typing.Any],
    ) -> None:
        """Take a player data delta and update the player data.

        Does not implement deleted as it is often overwritten with "modified".
        """
        recursively_delete_dict(self.data, delta["deleted"])
        recursively_update_dict(self.data, delta["modified"])

    async def request(self, endpoint: str, json: typing.Mapping[str, object] = {}, **kwargs: typing.Any) -> typing.Any:
        """Send an authenticated request to the arknights game server."""
        data = await super().request(endpoint, json=json, method="POST", **kwargs)
        self.update_player_data_delta(data["playerDataDelta"])
        if data.get("ts"):
            self.time_offset = data["ts"] - time.time()

        return data

    async def account_sync_data(self) -> typing.Any:
        """Sync player data.

        APP behavior: Called right after login.
        """
        data = await super().request("account/syncData", json={"platform": 1})
        self.data = data["user"]
        return data

    async def account_sync_status(self, modules: int = 95) -> typing.Any:
        """Sync progress status.

        I have literally no idea what modules means.
        I personally get 1759 after launch, 95 normally and sometimes 7.

        APP behavior: Called periodically after most actions.
        """
        data = {
            "modules": modules,
            "params": {
                "16": {"goodIdMap": {"LS": [], "HS": [], "ES": [], "CASH": [], "GP": ["GP_Once_1"], "SOCIAL": []}},
            },
        }
        return await self.request("account/syncStatus", json=data)

    async def user_check_in(self) -> typing.Any:
        """Claim daily login rewards.

        APP behavior: Called when claiming daily login rewards.
        """
        return await self.request("user/checkIn")

    async def mission_auto_confirm_missions(self, type: str) -> typing.Any:
        """Claim daily or weekly mission rewards,.

        type: Mission type. Either DAILY or WEEKLY.

        APP behavior: Called when batch completing daily or weekly mission rewards.
        """
        data = {
            "type": type,
        }
        return await self.request("mission/autoConfirmMissions", json=data)

    async def building_sync(self) -> typing.Any:
        """Sync base data.

        APP behavior: Called when entered the base.
        """
        return await self.request("building/sync")

    async def building_assign_char(self, room_slot_id: str, char_inst_id_list: typing.Sequence[int]) -> typing.Any:
        """Assign operators to a room.

        room_slot_id: Room slot ID to assign to.
        char_inst_id_list: List of operator instance IDs to assign. -1 for an empty slot.

        APP behavior: Called when assigning operators to a room.
        """
        data = {
            "roomSlotId": room_slot_id,
            "charInstIdList": char_inst_id_list,
        }
        return await self.request("building/assignChar", json=data)

    async def building_settle_manufacture(
        self,
        room_slot_id_list: typing.Sequence[str],
        supplement: bool = True,
    ) -> typing.Any:
        """Claim factory products.

        room_slot_id_list: List of room slot IDs to claim.
        supplement: Whether to supplement (restock) the materials back to the max.

        APP behavior: Called when claiming factory products.
        """
        data = {
            "roomSlotIdList": room_slot_id_list,
            "supplement": int(supplement),
        }
        return await self.request("building/settleManufacture", json=data)

    async def building_delivery_order(self, slot_id: str, order_id: int) -> typing.Any:
        """Claim trading post products.

        slot_id: Room slot ID to claim.
        order_id: Order ID to claim.

        APP behavior: Called when claiming individual trading post products.
        """
        data = {
            "slotId": slot_id,
            "orderId": order_id,
        }
        return await self.request("building/deliveryOrder", json=data)

    async def building_delivery_batch_order(self, slot_list: typing.Sequence[str]) -> typing.Any:
        """Claim all trading post products.

        slot_list: List of room slot IDs to claim.

        APP behavior: Called when claiming all trading post products.
        """
        data = {
            "slotList": slot_list,
        }
        return await self.request("building/deliveryBatchOrder", json=data)

    async def building_accelerate_solution(self, slot_id: str, cost: int) -> typing.Any:
        """Accelerate a room with drones.

        slot_id: Room slot ID to accelerate.
        cost: Amount of drones to use.

        APP behavior: Called when using drones.
        """
        data = {
            "slotId": slot_id,
            "cost": cost,
        }
        return await self.request("building/accelerateSolution", json=data)

    async def building_get_clue_box(self) -> typing.Any:
        """View clues gifted to your base by your friends.

        APP behavior: Called when opening the gifted clue box in the meeting room.
        """
        return await self.request("building/getClueBox")

    async def building_receive_clue_to_stock(self, clues: typing.Sequence[str]) -> typing.Any:
        """Claim clues gifted to your base by your friends.

        clues: List of clue IDs to claim.

        APP behavior: Called when claiming clues from the gifted clue box.
        """
        data = {
            "clues": clues,
        }
        return await self.request("building/receiveClueToStock", json=data)

    async def building_put_clue_to_the_board(self, clue_id: str) -> typing.Any:
        """Put a clue on the board or swap to the clue if there already is one.

        clue_id: Clue ID to put on the board.

        APP behavior: Called when putting a clue on the board.
        """
        data = {
            "clueId": clue_id,
        }
        return await self.request("building/putClueToTheBoard", json=data)

    async def building_take_clue_from_board(self, type: str) -> typing.Any:
        """Take a clue from the board.

        type: Clue type as a string. For example RHODES.

        APP behavior: Called when taking a clue from the board.
        """
        data = {
            "type": type,
        }
        return await self.request("building/takeClueFromBoard", json=data)

    async def building_visit_building(self, friend_id: str) -> typing.Any:
        """Visit a friend's base.

        friend_id: Friend's ID.

        APP behavior: Called when visiting a friend's base.
        """
        data = {
            "friendId": friend_id,
        }
        return await self.request("building/visitBuilding", json=data)

    async def building_get_assist_report(self) -> typing.Any:
        """Get base report.

        App behavior: Called when viewing base report in the control center.
        """
        return await self.request("building/getAssistReport")

    async def gacha_sync_normal_gacha(self) -> typing.Any:
        """Sync recruitment data.

        APP behavior: Called when entering the recruitment page.
        """
        return await self.request("gacha/syncNormalGacha")

    async def gacha_refresh_tags(self, slot_id: int) -> typing.Any:
        """Refresh recruitment tags.

        slot_id: Recruitment slot ID.

        APP behavior: Called when refreshing recruitment tags.
        """
        data = {
            "slotId": slot_id,
        }
        return await self.request("gacha/refreshTags", json=data)

    async def gacha_normal_gacha(
        self,
        slot_id: int,
        tag_list: typing.Sequence[int],
        duration: int,
        special_tag_id: int = -1,
    ) -> typing.Any:
        """Start a recruitment.

        slot_id: Recruitment slot ID.
        tag_list: List of tag IDs to use.
        duration: Recruitment duration in seconds.
        special_tag_id: Unknown, always -1.

        APP behavior: Called when starting a recruitment.
        """
        data = {
            "slotId": slot_id,
            "tagList": tag_list,
            "specialTagId": special_tag_id,
            "duration": duration,
        }
        return await self.request("gacha/normalGacha", json=data)

    async def gacha_finish_normal_gacha(self, slot_id: int) -> typing.Any:
        """Finish a recruitment.

        slot_id: Recruitment slot ID.

        APP behavior: Called when finishing a recruitment.
        """
        data = {
            "slotId": slot_id,
        }
        return await self.request("gacha/finishNormalGacha", json=data)

    # Maybe also get daily social points.

    async def shop_get_social_good_list(self) -> typing.Any:
        """Get the social shop product list.

        APP behavior: Called when entering the social shop.
        """
        return await self.request("shop/getSocialGoodList")

    async def shop_buy_social_good(self, good_id: str, count: int = 1) -> typing.Any:
        """Buy a social shop product.

        good_id: Product ID.
        count: Probably amount to buy. This should always be 1.

        APP behavior: Called when buying a social shop product.
        """
        data = {
            "goodId": good_id,
            "count": count,
        }
        return await self.request("shop/buySocialGood", json=data)

    async def get_battle_replay(self, battle_type: str, stage_id: str) -> typing.Any:
        """Get a replay of a stage.

        stage_id: Stage ID.

        APP behavior: Called when entering the squad selection screen with auto-deploy enabled.
        """
        data = {
            "stageId": stage_id,
        }
        return await self.request(f"{battle_type}/getBattleReplay", json=data)

    async def battle_start(
        self,
        battle_type: str,
        stage_id: str,
        squad: typing.Sequence[typing.Mapping[str, typing.Any]],
        use_practice_ticket: bool = False,
        is_replay: bool = True,
        pry: bool = False,
        is_retro: bool = False,
        assist_friend: typing.Any | None = None,
    ) -> typing.Any:
        """Start a stage.

        stage_id: Stage ID.
        squad: List of operators to deploy. Requires charInstId, currentEquip and skillIndex.
        use_practice_ticket: Whether to use a practice ticket.
        is_replay: Whether auto-deploy is being used.
        pry: Uknonwn, always false.
        is_retro: Unknown, always false.
        assist_friend: Unknown. None for no support.

        APP behavior: Called when starting a stage.
        """
        data = {
            "stageId": stage_id,
            "squad": squad,
            "usePracticeTicket": int(use_practice_ticket),
            "isReplay": int(is_replay),
            "pry": int(pry),
            "isRetro": int(is_retro),
            "assistFriend": assist_friend,
        }
        return await self.request(f"{battle_type}/battleStart", json=data)

    async def battle_finish(
        self,
        battle_type: str,
        stage_id: str,
        complete_time: int,
        is_cheat: str,
        stats: typing.Mapping[str, typing.Any],
        data: str,
    ) -> typing.Any:
        """Finish a stage. This is where you are most likely to get banned.

        stage_id: Stage ID.
        complete_time: Time taken to complete the stage in seconds.
        is_cheat: Encrypted battle id.
        stats: Unknown, always empty dict.
        data: Encrypted battle data.

        APP behavior: Called when finishing a stage.
        """
        d = {
            "battleData": {
                "stageId": stage_id,
                "completeTime": complete_time,
                "isCheat": is_cheat,
                "stats": stats,
            },
            "data": data,
        }
        return await self.request(f"{battle_type}/battleFinish", json=d)



"""Test arkprts client."""
import arkprts


def _force_forbid(cls: type[arkprts.models.BaseModel]) -> None:
    """Force forbid extra."""
    cls.model_config["extra"] = "forbid"
    for subclass in cls.__subclasses__():
        _force_forbid(subclass)


_force_forbid(arkprts.models.BaseModel)


async def test_search_players(client: arkprts.Client) -> None:
    players = await client.search_players("doctor", server="en", limit=10)
    assert len(players) == 10


async def test_get_data(client: arkprts.Client) -> None:
    # this is purposefully not allowed because it's useless for public clients
    data = await client.request("account/syncData", json={"platform": 1}, server="en")
    user = arkprts.models.User(client=client, **data["user"])
    assert user.status.nickname == "Doctor"  # default guest name



"""Test game data."""
import arkprts


async def test_download(client: arkprts.Client) -> None:
    # we could force, but that takes up a lot of time
    await client.update_gamedata()


def test_access(client: arkprts.Client) -> None:
    operator = client.gamedata.character_table["char_002_amiya"]
    assert operator.name == "Amiya"



"""Entry-point."""
import argparse
import asyncio
import logging

import arkprts

parser: argparse.ArgumentParser = argparse.ArgumentParser(description="Get user info from Arknights API.")
parser.add_argument("nickname", type=str, nargs="?", default=None, help="User nickname")
parser.add_argument("--uid", type=str, default=None, help="Channel UID")
parser.add_argument("--token", type=str, default="", help="Yostar Token")
parser.add_argument("--log-level", type=str, default="INFO", help="Logging level")
parser.add_argument("--server", type=str, default="en", help="Server to use, global only")


async def main() -> None:
    """Entry-point."""
    args = parser.parse_args()

    logging.basicConfig()
    logging.getLogger("arkprts").setLevel(args.log_level.upper())

    if args.uid and args.token:
        auth = arkprts.YostarAuth(args.server)
        await auth.login_with_token(args.uid, args.token)
    elif args.nickname:
        auth = arkprts.GuestAuth(max_sessions=1)
        auth.network.default_server = args.server
    else:
        auth = arkprts.YostarAuth(args.server)
        await auth.login_with_email_code()

    client = arkprts.Client(auth)

    if args.nickname:
        players = await client.search_players(args.nickname, limit=10)
    else:
        players = await client.get_friends()
        print("Friends:", end="\n\n")

    for player in players:
        if player.level < 5:
            continue

        print(f"{player.nickname}#{player.nick_number} ({player.uid}) Lvl {player.level}")
        print(f"Resume: {player.resume}")
        print(
            f"Current stage: {player.main_stage_progress} | Characters: {player.char_cnt} | Furniture: {player.furn_cnt} | "
            f"Secretary: {client.gamedata.character_table[player.secretary].name}",
        )
        print(f"Playing since: {player.register_ts.isoformat()}")
        print(f"Last Online: {player.last_online_time.isoformat()}")

        print("Support Operators: ", end="")
        for char in player.assist_char_list:
            if not char:
                continue
            print(f"{char.static.name} E{char.evolve_phase}L{char.level}", end="")
            if char.skills:
                print(f" S{char.skill_index+1}M{char.skills[char.skill_index].specialize_level}", end="")

            print("  ", end="")

        print("\n")


if __name__ == "__main__":
    asyncio.run(main())



"""Run setuptools."""
import pathlib

from setuptools import find_packages, setup

setup(
    name="arkprts",
    version="0.2.0",
    description="Arknights python wrapper.",
    url="https://github.com/thesadru/arkprts",
    packages=find_packages(exclude=["tests", "tests.*"]),
    include_package_data=True,
    package_data={"arkprts": ["py.typed"]},
    install_requires=["aiohttp", "pydantic==2.*"],
    extras_require={"all": ["rsa", "pycryptodome"], "rsa": ["rsa"], "aes": ["pycryptodome"]},
    long_description=pathlib.Path("README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    license="MIT",
)



"""Gamedata client. Fetches from github and such."""
from __future__ import annotations

import asyncio
import bisect
import fnmatch
import json
import logging
import os
import os.path
import pathlib
import subprocess
import tarfile
import tempfile
import time
import typing

import aiohttp

from arkprts.auth import ArknightsLanguage
from arkprts.models import base as models

__all__ = ("GameData",)

PathLike = typing.Union[pathlib.Path, str]

GITHUB_REPOSITORY = "aelurum/ArknightsGameData"  # zh-tw fork
FALLBACK_GITHUB_REPOSITORY = "Kengxxiao/ArknightsGameData"
RELEVANT_FILES = r"excel"

logger: logging.Logger = logging.getLogger("arkprts.gamedata")


class GameData:
    """Game data client."""

    directory: pathlib.Path
    """Directory path."""
    language: ArknightsLanguage
    """Default language."""
    loaded: bool
    """Whether the data was loaded at any point during the code execution."""

    _cache: dict[str, dict[str, models.DDict]]
    """Loaded files."""
    _lock: asyncio.Lock
    """Gamedata update lock."""

    def __init__(self, directory: PathLike | None = None, *, language: ArknightsLanguage = "en_US") -> None:
        if directory:
            self.directory = pathlib.Path(directory)
        else:
            self.directory = pathlib.Path(tempfile.gettempdir()) / "ArknightsGameData"

        self.directory.mkdir(parents=True, exist_ok=True)
        self.language = language
        self.loaded = False
        self._cache = {}
        self._lock = asyncio.Lock()

    @property
    def commit_file(self) -> pathlib.Path:
        """Commit path."""
        return self.directory / "commit.txt"

    async def _get_commit(self) -> str:
        """Get the commit hash."""
        url = f"https://api.github.com/repos/{GITHUB_REPOSITORY}/commits/master"
        async with aiohttp.ClientSession() as session, session.get(url) as response:
            response.raise_for_status()
            data = await response.json()
            return data["sha"]

    async def _download_tarball(self) -> None:
        """Download a tarball from github."""
        url = f"https://api.github.com/repos/{GITHUB_REPOSITORY}/tarball"
        async with aiohttp.ClientSession(auto_decompress=False) as session, session.get(url) as response:
            response.raise_for_status()
            with (self.directory / "ArknightsGameData.tar.gz").open("wb") as file:
                async for chunk in response.content.iter_any():
                    file.write(chunk)

    def _decompress_tarball(self, allow: str | None = None) -> str:
        """Decompress a tarball. Returns commit name."""
        allow = allow or RELEVANT_FILES

        with tarfile.open(self.directory / "ArknightsGameData.tar.gz") as tar:
            top_directory = os.path.commonprefix(tar.getnames())

            members: list[tarfile.TarInfo] = []
            for member in tar.getmembers():
                if not fnmatch.fnmatch(allow, member.name):
                    continue

                member.name = member.name[len(top_directory + "/") :]
                members.append(member)

            tar.extractall(self.directory, members=members)

        return top_directory.split("-")[-1]

    async def download_tarball_gamedata(self, allow: str | None = None, *, force: bool = False) -> None:
        """Download game data."""
        if not force and self.commit_file.exists() and time.time() - self.commit_file.stat().st_mtime < 60 * 60 * 6:
            logger.debug("Game data was updated recently, skipping download")
            return

        try:
            commit = await self._get_commit()
        except aiohttp.ClientResponseError:
            logger.warning("Failed to get game data commit, skipping download", exc_info=True)
            return

        if not force and self.commit_file.exists() and self.commit_file.read_text() == commit:
            logger.debug("Game data is up to date [%s]", commit)
            return

        logger.info("Downloading game data to %s [%s]", self.directory, commit)
        await self._download_tarball()

        logger.debug("Decompressing game data")
        tarball_commit = self._decompress_tarball(allow)
        logger.debug("Decompressed data %s", tarball_commit)
        if tarball_commit not in commit:
            raise RuntimeError(f"Tarball commit {tarball_commit} does not match github commit {commit}")

        # sometimes the contents are identical, we still want to update the mtime
        self.commit_file.write_text(commit)
        os.utime(self.commit_file, (time.time(), time.time()))

        logger.info("Downloaded game data")

    async def update_git_gamedata(self) -> None:
        """Update game data."""
        if not (self.directory / ".git").exists():
            logger.info("Initializing repository in %s", self.directory)
            proc = await asyncio.create_subprocess_exec(
                "git",
                "clone",
                "--depth=1",
                f"https://github.com/{GITHUB_REPOSITORY}.git",
                cwd=self.directory.parent,
            )
            await proc.wait()
        else:
            logger.info("Updating game data in %s", self.directory)
            proc = await asyncio.create_subprocess_exec("git", "pull", cwd=self.directory)
            await proc.wait()

    async def update_gamedata(self, allow: str | None = None, *, force: bool = False) -> None:
        """Update game data."""
        async with self._lock:
            proc = await asyncio.create_subprocess_exec(
                "git",
                "--version",
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            if (await proc.wait()) == 0:
                await self.update_git_gamedata()
            else:
                await self.download_tarball_gamedata(allow=allow, force=force)

            self.loaded = True

    def _get_data(self, filename: str, *, language: str | None = None) -> models.DDict:
        """Get a file."""
        language = language or self.language

        if self._cache.get(language, {}).get(filename):
            return models.DDict(self._cache[language][filename])

        path = self.directory / language / "gamedata" / filename
        if not path.exists():
            raise FileNotFoundError("Static gamedata has not been loaded.")

        with path.open(encoding="utf-8") as file:
            data = json.load(file)

        self._cache.setdefault(language, {})[filename] = data

        return models.DDict(data)

    def get_excel(self, name: str, *, language: str | None = None) -> models.DDict:
        """Get an excel table file."""
        return self._get_data(f"excel/{name}.json", language=language)

    def __getitem__(self, name: str) -> models.DDict:
        """Get an excel table file."""
        return self.get_excel(name)

    def __getattr__(self, name: str) -> models.DDict:
        """Get an excel table file."""
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(name) from e

    # helper stuff
    def get_operator(self, id: str, *, language: str | None = None) -> models.DDict:
        """Get an operator."""
        data = self.get_excel("character_table", language=language)
        return data[id]

    def get_item(self, id: str, *, language: str | None = None) -> models.DDict:
        """Get an item."""
        data = self.get_excel("item_table", language=language)
        return data["items"][id]

    def get_medal(self, id: str, *, language: str | None = None) -> models.DDict:
        """Get a medal."""
        data = self.get_excel("medal_table", language=language)
        return data["medalList"][id]

    def get_medal_group(self, id: str, *, language: str | None = None) -> models.DDict:
        """Get a medal group. Way too specific probably."""
        data = self.get_excel("medal_table", language=language)
        return next(i for i in data["medalTypeData"]["activityMedal"]["groupData"] if i["groupId"] == id)

    def get_skill(self, id: str, *, language: str | None = None) -> models.DDict:
        """Get a skill."""
        data = self.get_excel("skill_table", language=language)
        return data[id]

    def get_module(self, id: str, *, language: str | None = None) -> models.DDict:
        """Get a module."""
        data = self.get_excel("uniequip_table", language=language)
        return data["equipDict"][id]

    def calculate_trust_level(self, trust: int) -> int:
        """Calculate trust level from trust points."""
        frames = self.get_excel("favor_table")["favor_frames"]
        return bisect.bisect_left(frames, trust, key=lambda x: x["data"]["favor_point"])



"""Arknights client.

## Usage:
>>> # Client for read-only public data
>>> client = arkprts.Client()
>>> await client.search_players("...", limit=10)
[Player(...), ...]

>>> # Client for read-only private data
>>> auth = arkprts.YostarAuth("en")
>>> await auth.login_with_email_code("doctor@gmail.com")
>>> client = arkprts.Client(auth=auth)
>>> await client.get_data()
User(...)

>>> # Client for read & write (usage is potentially bannable)
>>> auth = arkprts.YostarAuth("en")
>>> await auth.login_with_email_code("doctor@gmail.com")
>>> client = arkprts.AutomationClient(auth=auth)
>>> await client.login_with_token("...", "...")
>>> await client.account_sync_data()
"""
from __future__ import annotations

import typing
import warnings

from . import auth as authn
from . import gamedata as gd
from . import models

if typing.TYPE_CHECKING:
    from typing_extensions import Self

__all__ = ["Client"]


class CoreClient:
    """Base arknights client."""

    auth: authn.CoreAuth
    """Authentication client."""
    gamedata: gd.GameData  # may actually be None, but that's a pain to typehint
    """Game data client."""

    def __init__(
        self,
        auth: authn.CoreAuth | None = None,
        *,
        gamedata: gd.GameData | typing.Literal[False] | None = None,
        network: authn.NetworkSession | None = None,
        server: authn.ArknightsServer | None = None,
        language: authn.ArknightsLanguage | None = None,
    ) -> None:
        """Initialize a client.

        auth: Authentication client. May be both public and private. GuestAuth by default.
        gamedata: Game data client. May be disabled with False.
        network: Network session.
        server: Default server. Not recommended for large-scale usage.
        language: Default language. Fallbacks on the gamedata's default language.
        """
        self.auth = auth or authn.GuestAuth(network=network)
        if gamedata is False:
            self.gamedata = None  # type: ignore
        else:
            self.gamedata = gamedata or gd.GameData()

        if network:
            self.auth.network = network
        if server:
            self.auth.network.default_server = server
        if language:
            if self.gamedata is None:
                raise ValueError("No need to use language, gamedata is disabled.")

            self.gamedata.language = language

    @property
    def network(self) -> authn.NetworkSession:
        """Return the network of the client."""
        return self.auth.network

    @property
    def server(self) -> authn.ArknightsServer | None:
        """Return the default server of the client."""
        return self.network.default_server

    async def request(self, endpoint: str, **kwargs: typing.Any) -> typing.Any:
        """Send an authenticated request to the arknights game server."""
        if self.gamedata and not self.gamedata.loaded:
            await self.gamedata.update_gamedata()

        return await self.auth.auth_request(endpoint, **kwargs)

    async def update_gamedata(self, allow: str | None = None, *, force: bool = False) -> bool:
        """Download game data."""
        if not self.gamedata:
            return False

        await self.gamedata.update_gamedata(allow=allow, force=force)
        return True

    @classmethod
    async def from_token(
        cls,
        channel_uid: str,
        token: str,
        server: authn.ArknightsServer = "en",
        *,
        network: authn.NetworkSession | None = None,
        gamedata: gd.GameData | None = None,
    ) -> Self:
        """Create a client from a token."""
        auth = await authn.Auth.from_token(server, channel_uid, token, network=network)
        return cls(auth, gamedata=gamedata)

    async def login_with_token(self, channel_uid: str, token: str) -> None:
        """Login with username and password."""
        warnings.warn(
            "client.login_with_token is deprecated, please use Client.from_token(...) or Client(auth=...)",
            category=DeprecationWarning,
        )
        self.auth = await authn.Auth.from_token(self.server or "en", channel_uid, token, network=self.network)


class Client(CoreClient):
    """Arknights client for accessing private data."""

    def _assert_private(self) -> None:
        """Assert that the client is not public."""
        if not isinstance(self.auth, authn.Auth):
            raise RuntimeError("This client can only access public data.")  # noqa: TRY004  # isn't a type check

        if not self.network.default_server:
            raise RuntimeError("Missing a default server for a private client.")

    async def get_raw_data(self) -> typing.Any:
        """Get user data."""
        self._assert_private()

        return await self.request("account/syncData", json={"platform": 1})

    async def _get_social_sort_list(
        self,
        type: int,
        sort_key: typing.Sequence[str] = ["level"],
        param: typing.Mapping[str, str] = {},
        *,
        server: authn.ArknightsServer | None = None,
    ) -> typing.Any:
        """Request sortedusers."""
        data = await self.request(
            "social/getSortListInfo",
            json={"type": type, "sortKeyList": sort_key, "param": param},
            server=server,
        )
        data["result"].sort(key=lambda x: tuple(x[key] for key in sort_key), reverse=True)  # pyright: ignore

        return data

    async def get_raw_friend_info(
        self,
        ids: typing.Sequence[str],
        *,
        server: authn.ArknightsServer | None = None,
    ) -> typing.Any:
        """Get detailed player info. You don't need to be friends actually."""
        return await self.request("social/getFriendList", json={"idList": ids}, server=server)

    async def get_raw_player_info(
        self,
        ids: typing.Sequence[str],
        *,
        server: authn.ArknightsServer | None = None,
    ) -> typing.Any:
        """Get player info."""
        return await self.request("social/searchPlayer", json={"idList": ids}, server=server)

    async def get_raw_friend_ids(
        self,
        *,
        server: authn.ArknightsServer | None = None,
    ) -> typing.Any:
        """Get friends."""
        self._assert_private()

        return await self._get_social_sort_list(1, ["level", "infoShare"], {}, server=server)

    async def search_raw_player_ids(
        self,
        nickname: str,
        nicknumber: str = "",
        *,
        server: authn.ArknightsServer | None = None,
    ) -> typing.Any:
        """Search for a nickname."""
        return await self._get_social_sort_list(
            0,
            ["level"],
            {"nickName": nickname, "nickNumber": nicknumber},
            server=server,
        )

    async def search_players(
        self,
        nickname: str,
        nicknumber: str = "",
        *,
        server: authn.ArknightsServer | None = None,
        limit: int | None = None,
    ) -> typing.Sequence[models.Player]:
        """Search for a player and return a model."""
        if "#" in nickname:
            nickname, nicknumber = nickname.split("#", 1)

        uid_data = await self.search_raw_player_ids(nickname, nicknumber, server=server)
        data = await self.get_raw_friend_info([uid["uid"] for uid in uid_data["result"][:limit]], server=server)
        return [models.Player(client=self, **i) for i in data["friends"]]

    async def get_players(
        self,
        ids: typing.MutableSequence[str],
        *,
        server: authn.ArknightsServer | None = None,
    ) -> typing.Sequence[models.Player]:
        """Get players and return a model."""
        data = await self.get_raw_player_info(ids, server=server)
        return [models.Player(client=self, **i) for i in data["result"]]

    async def get_friends(
        self,
        *,
        server: authn.ArknightsServer | None = None,
        limit: int | None = None,
    ) -> typing.Sequence[models.Player]:
        """Get friends and return a model."""
        uid_data = await self.get_raw_friend_ids(server=server)
        data = await self.get_raw_friend_info([uid["uid"] for uid in uid_data["result"][:limit]], server=server)
        return [models.Player(client=self, **i) for i in data["friends"]]

    async def get_data(self) -> models.User:
        """Get user sync data and return a model. Use raw data for more info."""
        data = await self.get_raw_data()
        return models.User(client=self, **data["user"])



"""Modified pydantic base model."""
import collections
import datetime
import typing

import pydantic
import pydantic_core

from arkprts.client import CoreClient

__all__ = ("BaseModel", "DDict")

# pydantic hack
_fake_client = type("", (object,), {})()
_fake_client.__class__ = CoreClient


def _set_recursively(obj: typing.Any, name: str, value: typing.Any) -> None:
    """Set an attribute recursively."""
    if isinstance(obj, BaseModel):
        object.__setattr__(obj, name, value)
        for field in type(obj).model_fields:
            _set_recursively(getattr(obj, field), name, value)

    elif isinstance(obj, typing.Mapping):
        for key, item in obj.items():  # pyright: ignore[reportUnknownVariableType]
            _set_recursively(key, name, value)
            _set_recursively(item, name, value)
    elif isinstance(obj, typing.Sequence) and not isinstance(obj, str):
        for item in obj:  # pyright: ignore[reportUnknownVariableType]
            _set_recursively(item, name, value)


def _to_camel_case(string: str) -> str:
    """Convert snake_case to camelCase."""
    return "".join(x.title() if i else x for i, x in enumerate(string.split("_")))


def parse_timestamp(timestamp: int) -> datetime.datetime | None:
    """Parse an arknights timestamp and use the local timezone.

    Arknights provides timestamps not with UTC but with the client's timezone.
    """
    if timestamp in (0, -1):
        return None

    return datetime.datetime.fromtimestamp(timestamp).astimezone()  # noqa: DTZ006


class BaseModel(pydantic.BaseModel, arbitrary_types_allowed=True):
    """Client-aware pydantic base model."""

    client: CoreClient = pydantic.Field(repr=False)
    """Client instance."""

    def __init__(self, client: typing.Optional[CoreClient] = None, **kwargs: typing.Any) -> None:
        """Init."""
        super().__init__(client=_fake_client, **kwargs)
        if client:
            _set_recursively(self, "client", client)

    @pydantic.model_validator(mode="before")
    def _fix_amiya(cls, value: typing.Any, info: pydantic.ValidationInfo) -> typing.Any:
        """Flatten Amiya to only keep her selected form if applicable."""
        if value and value.get("tmpl"):
            current = value["tmpl"][value["currentTmpl"]]
            value.update(current)

        return value


class DList(collections.UserList[typing.Any]):
    """Dot-accessed list."""

    def __getitem__(self, key: typing.Any) -> typing.Any:
        item = super().__getitem__(key)

        if isinstance(item, dict):
            item = DDict(item)  # pyright: ignore[reportUnknownArgumentType]
        elif isinstance(item, list):
            item = DList(item)  # pyright: ignore[reportUnknownArgumentType]

        return item

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source: typing.Any,
        handler: typing.Callable[[typing.Any], pydantic_core.CoreSchema],
    ) -> pydantic_core.CoreSchema:
        """Generate a pydantic core schema."""
        return pydantic_core.core_schema.no_info_plain_validator_function(cls)


class DDict(collections.UserDict[str, typing.Any]):
    """Dot-accessed dictionary."""

    def __getitem__(self, key: typing.Any) -> typing.Any:
        try:
            item = super().__getitem__(key)
        except KeyError:
            item = super().__getitem__(_to_camel_case(key))

        if isinstance(item, dict):
            item = DDict(item)  # pyright: ignore[reportUnknownArgumentType]
        elif isinstance(item, list):
            item = DList(item)  # pyright: ignore[reportUnknownArgumentType]

        return item

    def __getattr__(self, key: str) -> typing.Any:
        try:
            return self[key]
        except KeyError as e:
            raise AttributeError(*e.args) from e

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source: typing.Any,
        handler: typing.Callable[[typing.Any], pydantic_core.CoreSchema],
    ) -> pydantic_core.CoreSchema:
        """Generate a pydantic core schema."""
        return pydantic_core.core_schema.no_info_plain_validator_function(cls)


ArknightsTimestamp = typing.Annotated[datetime.datetime, pydantic.PlainValidator(parse_timestamp)]
